@@ -1,6 +1,5 @@
-/* gda-data-comparator.c
- *
- * Copyright (C) 2008 Vivien Malerba
+/*
+ * Copyright (C) 2008 - 2011 Vivien Malerba
  *
  * This Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -98,7 +97,8 @@ gda_data_comparator_get_type (void)
 			NULL,
 			sizeof (GdaDataComparator),
 			0,
-			(GInstanceInitFunc) gda_data_comparator_init
+			(GInstanceInitFunc) gda_data_comparator_init,
+			0
 		};
 
 		g_static_mutex_lock (&registering);
@@ -110,10 +110,10 @@ gda_data_comparator_get_type (void)
 }
 
 static gboolean
-diff_computed_accumulator (GSignalInvocationHint *ihint,
+diff_computed_accumulator (G_GNUC_UNUSED GSignalInvocationHint *ihint,
 			   GValue *return_accu,
 			   const GValue *handler_return,
-			   gpointer data)
+			   G_GNUC_UNUSED gpointer data)
 {
         gboolean thisvalue;
 
@@ -124,7 +124,7 @@ diff_computed_accumulator (GSignalInvocationHint *ihint,
 }
 
 static gboolean
-m_diff_computed (GdaDataComparator *comparator, GdaDiff *diff)
+m_diff_computed (G_GNUC_UNUSED GdaDataComparator *comparator, G_GNUC_UNUSED GdaDiff *diff)
 {
         return FALSE; /* default is to allow differences computing to proceed (understand it as: FALSE => don't stop) */
 }
@@ -174,7 +174,7 @@ gda_data_comparator_init (GdaDataComparator *comparator)
 }
 
 /**
- * gda_data_comparator_new
+ * gda_data_comparator_new:
  * @old_model: Data model to which the modifications should be applied
  * @new_model: Target data model.
  *
@@ -182,7 +182,7 @@ gda_data_comparator_init (GdaDataComparator *comparator)
  * all the computed differences (as #GdaDiff structures) to @old_model, the resulting data model
  * should have the same contents as @new_model.
  *
- * Returns: a new #GdaDataComparator object
+ * Returns: (type GdaDataComparator) (transfer full): a new #GdaDataComparator object
  */
 GObject *
 gda_data_comparator_new (GdaDataModel *old_model, GdaDataModel *new_model)
@@ -201,7 +201,7 @@ static void
 clean_diff (GdaDataComparator *comparator)
 {
 	if (comparator->priv->diffs) {
-		gint i;
+		gsize i;
 		for (i = 0; i < comparator->priv->diffs->len; i++) {
 			GdaDiff *diff = g_array_index (comparator->priv->diffs, GdaDiff *, i);
 			gda_diff_free (diff);
@@ -294,6 +294,9 @@ gda_data_comparator_set_property (GObject *object,
 			if (model)
 				g_object_ref (model);
 			break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
 		}
 	}
 }
@@ -315,6 +318,9 @@ gda_data_comparator_get_property (GObject *object,
 		case PROP_NEW_MODEL:
 			g_value_set_object (value, comparator->priv->new_model);
 			break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
 		}
 	}
 }
@@ -327,10 +333,10 @@ gda_diff_free (GdaDiff *diff)
 }
 
 /**
- * gda_data_comparator_set_key_columns
+ * gda_data_comparator_set_key_columns:
  * @comp: a #GdaDataComparator object
  * @nb_cols: the size of the @col_numbers array
- * @col_numbers: a array of @nb_cols values
+ * @col_numbers: (array length=nb_cols): an array of @nb_cols values
  *
  * Defines the columns which will be used as a key when searching data. This is not mandatory but
  * will speed things up as less data will be processed.
@@ -345,8 +351,8 @@ gda_data_comparator_set_key_columns (GdaDataComparator *comp, const gint *col_nu
 	comp->priv->key_columns = NULL;
 	if (nb_cols > 0) {
 		comp->priv->nb_key_columns = nb_cols;
-		comp->priv->key_columns = g_new0 (gint, nb_cols);
-		memcpy (comp->priv->key_columns, col_numbers, sizeof (gint) * nb_cols);
+		comp->priv->key_columns = g_new (gint, nb_cols);
+		memcpy (comp->priv->key_columns, col_numbers, sizeof (gint) * nb_cols); /* Flawfinder: ignore */
 	}
 }
 
@@ -411,11 +417,11 @@ find_row_in_model (GdaDataComparator *comp, gint row, gboolean *out_has_changed,
 }
 
 /**
- * gda_data_comparator_compute_diff
+ * gda_data_comparator_compute_diff:
  * @comp: a #GdaDataComparator object
  * @error: a place to store errors, or %NULL
  *
- * Actually computes the differences bewteen the data models for which @comp is defined. 
+ * Actually computes the differences between the data models for which @comp is defined. 
  *
  * For each difference computed, stored in a #GdaDiff structure, the "diff-computed" signal is emitted.
  * If one connects to this signal and returns FALSE in the signal handler, then computing differences will be
@@ -478,11 +484,21 @@ gda_data_comparator_compute_diff (GdaDataComparator *comp, GError **error)
 
 	/* actual differences computations : rows to insert / update */
 	onrows = gda_data_model_get_n_rows (comp->priv->old_model);
-	g_assert (onrows >= 0);
+	if (onrows < 0) {
+		g_set_error (error, GDA_DATA_COMPARATOR_ERROR, GDA_DATA_COMPARATOR_MODEL_ACCESS_ERROR,
+			     "%s", _("Can't get the number of rows of data model to compare from"));
+		return FALSE;
+	}
 	nnrows = gda_data_model_get_n_rows (comp->priv->new_model);
-	g_assert (nnrows >= 0);
-	rows_to_del = g_new (gboolean, onrows);
-	memset (rows_to_del, TRUE, sizeof (gboolean) * onrows);
+	if (nnrows < 0) {
+		g_set_error (error, GDA_DATA_COMPARATOR_ERROR, GDA_DATA_COMPARATOR_MODEL_ACCESS_ERROR,
+			     "%s", _("Can't get the number of rows of data model to compare to"));
+		return FALSE;
+	}
+	if (onrows > 0) {
+		rows_to_del = g_new (gboolean, onrows);
+		memset (rows_to_del, TRUE, sizeof (gboolean) * onrows);
+	}
 	for (i = 0; i < nnrows; i++) {
 		gint erow = -1;
 		gboolean has_changed = FALSE;
@@ -525,6 +541,7 @@ gda_data_comparator_compute_diff (GdaDataComparator *comp, GError **error)
 			gint j;
 			diff = g_new0 (GdaDiff, 1);
 			diff->type = GDA_DIFF_MODIFY_ROW;
+			rows_to_del [erow] = FALSE;
 			diff->old_row = erow;
 			diff->new_row = i;
 			diff->values = g_hash_table_new_full (g_str_hash, g_str_equal, (GDestroyNotify) g_free,
@@ -551,7 +568,9 @@ gda_data_comparator_compute_diff (GdaDataComparator *comp, GError **error)
 						     gda_value_copy (cvalue));
 			}
 		}
-		rows_to_del [i] = FALSE;
+		else
+			rows_to_del [erow] = FALSE; /* row has not been changed */
+
 		if (diff) {
 			g_array_append_val (comp->priv->diffs, diff);
 			g_signal_emit (comp, gda_data_comparator_signals [DIFF_COMPUTED], 0, diff, &stop);
@@ -572,7 +591,7 @@ gda_data_comparator_compute_diff (GdaDataComparator *comp, GError **error)
 		if (rows_to_del [i]) {
 			gint j;
 			diff = g_new0 (GdaDiff, 1);
-			diff->type = GDA_DIFF_ADD_ROW;
+			diff->type = GDA_DIFF_REMOVE_ROW;
 			diff->old_row = i;
 			diff->new_row = -1;
 			diff->values = g_hash_table_new_full (g_str_hash, g_str_equal, (GDestroyNotify) g_free,
@@ -606,7 +625,7 @@ gda_data_comparator_compute_diff (GdaDataComparator *comp, GError **error)
 }
 
 /**
- * gda_data_comparator_get_n_diffs
+ * gda_data_comparator_get_n_diffs:
  * @comp: a #GdaDataComparator object
  *
  * Get the number of differences as computed by the last time gda_data_comparator_compute_diff() was called.
@@ -623,13 +642,13 @@ gda_data_comparator_get_n_diffs  (GdaDataComparator *comp)
 }
 
 /**
- * gda_data_comparator_get_diff
+ * gda_data_comparator_get_diff:
  * @comp: a #GdaDataComparator object
  * @pos: the requested difference number (starting at 0)
  *
  * Get a pointer to the #GdaDiff structure representing the difference which number is @pos
  *
- * Returns: a pointer to a #GdaDiff, or %NULL if @pos is invalid
+ * Returns: (transfer none): a pointer to a #GdaDiff, or %NULL if @pos is invalid
  */
 const GdaDiff *
 gda_data_comparator_get_diff (GdaDataComparator *comp, gint pos)
