@@ -1,9 +1,10 @@
-/* GDA library
- * Copyright (C) 1998 - 2009 The GNOME Foundation.
+/*
+ * Copyright (C) 1998 - 2011 The GNOME Foundation.
  *
  * AUTHORS:
  *	Rodrigo Moya <rodrigo@gnome-db.org>
  *      Vivien Malerba <malerba@gnome-db.org>
+ *      Daniel Espinosa <esodan@gmail.com>
  *
  * This Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -125,7 +126,8 @@ gda_data_model_array_get_type (void)
 			NULL,
 			sizeof (GdaDataModelArray),
 			0,
-			(GInstanceInitFunc) gda_data_model_array_init
+			(GInstanceInitFunc) gda_data_model_array_init,
+			0
 		};
 
 		static const GInterfaceInfo data_model_info = {
@@ -173,14 +175,14 @@ gda_data_model_array_class_init (GdaDataModelArrayClass *klass)
 }
 
 static void
-gda_data_model_array_init (GdaDataModelArray *model, GdaDataModelArrayClass *klass)
+gda_data_model_array_init (GdaDataModelArray *model, G_GNUC_UNUSED GdaDataModelArrayClass *klass)
 {
 	g_return_if_fail (GDA_IS_DATA_MODEL_ARRAY (model));
 
 	/* allocate internal structure */
 	model->priv = g_new0 (GdaDataModelArrayPrivate, 1);
 	model->priv->notify_changes = TRUE;
-        model->priv->column_spec = g_hash_table_new (g_direct_hash, g_direct_equal);
+        model->priv->column_spec = g_hash_table_new_full (g_int_hash, g_int_equal, g_free, NULL);
         model->priv->read_only = FALSE;
 	model->priv->number_of_columns = 0;
 	model->priv->rows = g_array_new (FALSE, FALSE, sizeof (GdaRow *));
@@ -189,14 +191,13 @@ gda_data_model_array_init (GdaDataModelArray *model, GdaDataModelArrayClass *kla
 static void column_g_type_changed_cb (GdaColumn *column, GType old, GType new, GdaDataModelArray *model);
 
 static void
-hash_free_column (gpointer key, GdaColumn *column, GdaDataModelArray *model)
+hash_free_column (G_GNUC_UNUSED gpointer key, GdaColumn *column, GdaDataModelArray *model)
 {
         g_signal_handlers_disconnect_by_func (G_OBJECT (column),
                                               G_CALLBACK (column_g_type_changed_cb), model);
         g_object_unref (column);
 }
 
-
 static void
 gda_data_model_array_finalize (GObject *object)
 {
@@ -264,13 +265,13 @@ gda_data_model_array_get_property (GObject *object, guint prop_id, GValue *value
 
 
 /**
- * gda_data_model_array_new
+ * gda_data_model_array_new:
  * @cols: number of columns for rows in this data model.
  *
  * Creates a new #GdaDataModel object without initializing the column
  * types. Using gda_data_model_array_new_with_g_types() is usually better.
  *
- * Returns: a pointer to the newly created #GdaDataModel.
+ * Returns: (transfer full): a pointer to the newly created #GdaDataModel.
  */
 GdaDataModel *
 gda_data_model_array_new (gint cols)
@@ -282,14 +283,14 @@ gda_data_model_array_new (gint cols)
 }
 
 /**
- * gda_data_model_array_new_with_g_types
+ * gda_data_model_array_new_with_g_types: (skip)
  * @cols: number of columns for rows in this data model.
  * @...: types of the columns of the model to create as #GType, as many as indicated by @cols
  * 
  * Creates a new #GdaDataModel object with the column types as
  * specified.
  *
- * Returns: a pointer to the newly created #GdaDataModel.
+ * Returns: (transfer full): a pointer to the newly created #GdaDataModel.
  */
 GdaDataModel *
 gda_data_model_array_new_with_g_types (gint cols, ...)
@@ -316,13 +317,41 @@ gda_data_model_array_new_with_g_types (gint cols, ...)
 }
 
 /**
- * gda_data_model_array_copy_model
+ * gda_data_model_array_new_with_g_types_v:
+ * @cols: number of columns for rows in this data model.
+ * @types: array of types of the columns of the model to create as #GType, as many as indicated by @cols
+ *
+ * Creates a new #GdaDataModel object with the column types as
+ * specified.
+ *
+ * Returns: (transfer full): a pointer to the newly created #GdaDataModel.
+ *
+ * Since: 4.2.6
+ */
+GdaDataModel *
+gda_data_model_array_new_with_g_types_v (gint cols, GType *types)
+{
+	GdaDataModel *model;
+	gint i;
+
+	model = gda_data_model_array_new (cols);
+	i = 0;
+	while (i < cols) {
+		GType type = types [i];
+		gda_column_set_g_type (gda_data_model_describe_column (model, i), type);
+		i++;
+	}
+	return model;
+}
+
+/**
+ * gda_data_model_array_copy_model:
  * @src: a #GdaDataModel to copy data from
  * @error: a place to store errors, or %NULL
  *
  * Makes a copy of @src into a new #GdaDataModelArray object
  *
- * Returns: a new data model, or %NULL if an error occurred
+ * Returns: (transfer full) (allow-none): a new data model, or %NULL if an error occurred
  */
 GdaDataModelArray *
 gda_data_model_array_copy_model (GdaDataModel *src, GError **error)
@@ -368,21 +397,22 @@ gda_data_model_array_copy_model (GdaDataModel *src, GError **error)
 }
 
 /**
- * gda_data_model_array_get_row
+ * gda_data_model_array_get_row:
  * @model: a #GdaDataModelArray object
  * @row: row number (starting from 0)
  * @error: a place to store errors, or %NULL
  *
  * Get a pointer to a row in @model
  *
- * Returns: the #GdaRow, or %NULL if an error occurred
+ * Returns: (transfer none): the #GdaRow, or %NULL if an error occurred
  */
 GdaRow *
 gda_data_model_array_get_row (GdaDataModelArray *model, gint row, GError **error)
 {
 	g_return_val_if_fail (GDA_IS_DATA_MODEL_ARRAY (model), NULL);
+	g_return_val_if_fail (row >= 0, NULL);
 
-	if (row >= model->priv->rows->len) {
+	if ((guint)row >= model->priv->rows->len) {
 		if (model->priv->rows->len > 0)
 			g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_ROW_OUT_OF_RANGE_ERROR,
 				     _("Row %d out of range (0-%d)"), row,
@@ -397,7 +427,7 @@ gda_data_model_array_get_row (GdaDataModelArray *model, gint row, GError **error
 }
 
 /**
- * gda_data_model_array_set_n_columns
+ * gda_data_model_array_set_n_columns:
  * @model: the #GdaDataModelArray.
  * @cols: number of columns for rows this data model should use.
  *
@@ -418,7 +448,7 @@ gda_data_model_array_set_n_columns (GdaDataModelArray *model, gint cols)
 }
 
 /**
- * gda_data_model_array_clear
+ * gda_data_model_array_clear:
  * @model: the model to clear.
  *
  * Frees all the rows in @model.
@@ -460,29 +490,33 @@ gda_data_model_array_describe_column (GdaDataModel *model, gint col)
                 return NULL;
         }
 
-        column = g_hash_table_lookup (((GdaDataModelArray*) model)->priv->column_spec,
-                                      GINT_TO_POINTER (col));
+	gint tmp;
+	tmp = col;
+        column = g_hash_table_lookup (((GdaDataModelArray*) model)->priv->column_spec, &tmp);
         if (!column) {
                 column = gda_column_new ();
                 g_signal_connect (G_OBJECT (column), "g-type-changed",
                                   G_CALLBACK (column_g_type_changed_cb), model);
                 gda_column_set_position (column, col);
-                g_hash_table_insert (((GdaDataModelArray*) model)->priv->column_spec,
-                                     GINT_TO_POINTER (col), column);
+
+		gint *ptr;
+		ptr = g_new (gint, 1);
+		*ptr = col;
+                g_hash_table_insert (((GdaDataModelArray*) model)->priv->column_spec, ptr, column);
         }
 
         return column;
 }
 
 static void
-column_g_type_changed_cb (GdaColumn *column, GType old, GType new, GdaDataModelArray *model)
+column_g_type_changed_cb (GdaColumn *column, G_GNUC_UNUSED GType old, GType new, GdaDataModelArray *model)
 {
         /* emit a warning if there are GValues which are not compatible with the new type */
         gint i, nrows, col;
         const GValue *value;
         gchar *str;
         gint nb_warnings = 0;
-#define max_warnings 5
+	const gint max_warnings = 5;
 
         if ((new == G_TYPE_INVALID) ||
             (new == GDA_TYPE_NULL))
@@ -523,13 +557,15 @@ gda_data_model_array_get_value_at (GdaDataModel *model, gint col, gint row, GErr
 	GdaRow *fields;
 	GdaDataModelArray *amodel = (GdaDataModelArray*) model;
 
+	g_return_val_if_fail(row >= 0, NULL);
+
 	if (amodel->priv->rows->len == 0) {
 		g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_ROW_NOT_FOUND_ERROR,
 			      "%s", _("No row in data model"));
 		return NULL;
 	}
 
-	if (row >= amodel->priv->rows->len) {
+	if ((guint)row >= amodel->priv->rows->len) {
 		if (amodel->priv->rows->len > 0)
 			g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_ROW_OUT_OF_RANGE_ERROR,
 				     _("Row %d out of range (0-%d)"), row, amodel->priv->rows->len - 1);
@@ -616,7 +652,7 @@ gda_data_model_array_set_value_at (GdaDataModel *model, gint col, gint row,
                 return FALSE;
         }
 
-	if (row > amodel->priv->rows->len) {
+	if ((guint)row > amodel->priv->rows->len) {
 		if (amodel->priv->rows->len > 0)
 			g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_VALUES_LIST_ERROR,
 				     _("Row %d out of range (0-%d)"), row, amodel->priv->rows->len - 1);
@@ -660,7 +696,7 @@ gda_data_model_array_set_values (GdaDataModel *model, gint row, GList *values, G
                 return FALSE;
         }
 
-        if (g_list_length (values) > gda_data_model_get_n_columns (model)) {
+        if (g_list_length (values) > (guint)gda_data_model_get_n_columns (model)) {
                 g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_VALUES_LIST_ERROR,
                               "%s", _("Too many values in list"));
                 return FALSE;
@@ -699,7 +735,7 @@ gda_data_model_array_append_values (GdaDataModel *model, const GList *values, GE
                 return FALSE;
         }
 
-	if (g_list_length ((GList *) values) > amodel->priv->number_of_columns) {
+	if (g_list_length ((GList *) values) > (guint)amodel->priv->number_of_columns) {
                 g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_VALUES_LIST_ERROR,
                               "%s", _("Too many values in list"));
                 return FALSE;
