@@ -1,6 +1,5 @@
 /* 
- * GDA common library
- * Copyright (C) 1998 - 2010 The GNOME Foundation.
+ * Copyright (C) 1998 - 2011 The GNOME Foundation.
  *
  * AUTHORS:
  *	Rodrigo Moya <rodrigo@gnome-db.org>
@@ -39,6 +38,7 @@
 #include <string.h>
 #ifdef HAVE_LOCALE_H
 #ifndef G_OS_WIN32
+#include <sys/ioctl.h>
 #include <langinfo.h>
 #endif
 #include <locale.h>
@@ -771,7 +771,9 @@ gda_data_model_set_values (GdaDataModel *model, gint row, GList *values, GError
  * @model: a #GdaDataModel object.
  *
  * Creates a new iterator object #GdaDataModelIter object which can be used to iterate through
- * rows in @model.
+ * rows in @model. The new #GdaDataModelIter does not hold any reference to @model (ie. if @model
+ * is destroyed at some point, the new iterator will become useless but in any case it will not prevent
+ * the data model from being destroyed).
  *
  * Depending on the data model's implementation, a new #GdaDataModelIter object may be created,
  * or a reference to an already existing #GdaDataModelIter may be returned.
@@ -994,6 +996,31 @@ gda_data_model_send_hint (GdaDataModel *model, GdaDataModelHint hint, const GVal
 		(GDA_DATA_MODEL_GET_CLASS (model)->i_send_hint) (model, hint, hint_value);
 }
 
+/**
+ * gda_data_model_get_exceptions:
+ * @model: a #GdaDataModel
+ *
+ * Get the global data model exception(s) that occurred when using @model.
+ * This is usefull for example for the LDAP related
+ * data models where some rows may be missing because the LDAP search has reached a limit
+ * imposed by the LDAP server.
+ *
+ * Returns: (transfer none) (element-type GError) (array zero-terminated=1): a pointer to a %NULL terminated array of #GError, or %NULL.
+ *
+ * Since: 4.2.6
+ */
+GError **
+gda_data_model_get_exceptions (GdaDataModel *model)
+{
+	g_return_val_if_fail (GDA_IS_DATA_MODEL (model), NULL);
+
+	if (GDA_DATA_MODEL_GET_CLASS (model)->i_get_exceptions)
+		return (GDA_DATA_MODEL_GET_CLASS (model)->i_get_exceptions) (model);
+	else
+		return NULL;
+}
+
+
 static gchar *export_to_text_separated (GdaDataModel *model, const gint *cols, gint nb_cols, 
 					const gint *rows, gint nb_rows, gchar sep, gchar quote, gboolean field_quotes,
 					gboolean null_as_empty, gboolean invalid_as_null);
@@ -1699,7 +1726,16 @@ gda_data_model_add_data_from_xml_node (GdaDataModel *model, xmlNodePtr node, GEr
  *
  * The @cols_trans is a hash table for which keys are @to columns numbers and the values are
  * the corresponding column numbers in the @from data model. To set the values of a column in @to to NULL,
- * create an entry in the hash table with a negative value.
+ * create an entry in the hash table with a negative value. For example:
+ * <programlisting><![CDATA[GHashTable *hash;
+ * gint *ptr;
+ * hash = g_hash_table_new_full (g_int_hash, g_int_equal, g_free, NULL);
+ * ptr = g_new (gint, 1);
+ * *ptr = 2;
+ * g_hash_table_insert (hash, ptr, GINT_TO_POINTER (3));
+ * gda_data_model_import_from_model (...);
+ * g_hash_table_free (hash);
+ * ]]></programlisting>
  *
  * Upon errors FALSE will be returned and @error will be assigned a
  * #GError from the #GDA_DATA_MODEL_ERROR domain.
@@ -1748,8 +1784,8 @@ gda_data_model_import_from_model (GdaDataModel *to, GdaDataModel *from,
 		GdaColumn *column;
 
 		if (cols_trans) {
-			col = GPOINTER_TO_INT (g_hash_table_lookup (cols_trans, GINT_TO_POINTER (i)));
-			if (col >= from_nb_cols) {
+			col = GPOINTER_TO_INT (g_hash_table_lookup (cols_trans, &i));
+			if ((col < 0) || (col >= from_nb_cols)) {
 				g_slist_free (copy_params);
 				g_set_error (error, GDA_DATA_MODEL_ERROR, GDA_DATA_MODEL_COLUMN_OUT_OF_RANGE_ERROR,
 					     _("Inexistent column in source data model: %d"), col);
@@ -2016,7 +2052,7 @@ gda_data_model_import_from_file (GdaDataModel *model,
 
 static gchar *real_gda_data_model_dump_as_string (GdaDataModel *model, gboolean dump_attributes, 
 						  gboolean dump_rows, gboolean dump_title, gboolean null_as_empty,
-						  GError **error);
+						  gint max_width, GError **error);
 
 /**
  * gda_data_model_dump:
@@ -2029,8 +2065,9 @@ static gchar *real_gda_data_model_dump_as_string (GdaDataModel *model, gboolean
  * <itemizedlist>
  *   <listitem><para>GDA_DATA_MODEL_DUMP_ROW_NUMBERS: if set, the first column of the output will contain row numbers</para></listitem>
  *   <listitem><para>GDA_DATA_MODEL_DUMP_ATTRIBUTES: if set, also dump the data model's columns' types and value's attributes</para></listitem>
-*   <listitem><para>GDA_DATA_MODEL_DUMP_TITLE: if set, also dump the data model's title</para></listitem>
-*   <listitem><para>GDA_DATA_MODEL_DUMP_NULL_AS_EMPTY: if set, replace the 'NULL' string with an empty string for NULL values </para></listitem>
+ *   <listitem><para>GDA_DATA_MODEL_DUMP_TITLE: if set, also dump the data model's title</para></listitem>
+ *   <listitem><para>GDA_DATA_MODEL_NULL_AS_EMPTY: if set, replace the 'NULL' string with an empty string for NULL values </para></listitem>
+ *   <listitem><para>GDA_DATA_MODEL_DUMP_TRUNCATE: if set to a numeric value, truncates the output to the width specified by the value. If the value is -1 then the actual terminal size (if it can be determined) is used</para></listitem>
  * </itemizedlist>
  */
 void
@@ -2041,7 +2078,9 @@ gda_data_model_dump (GdaDataModel *model, FILE *to_stream)
 	gboolean dump_rows = FALSE;
 	gboolean dump_title = FALSE;
 	gboolean null_as_empty = FALSE;
+	gint max_width = 0; /* no truncate */
 	GError *error = NULL;
+	const gchar *cstr;
 
 	g_return_if_fail (GDA_IS_DATA_MODEL (model));
 	if (!to_stream)
@@ -2055,13 +2094,26 @@ gda_data_model_dump (GdaDataModel *model, FILE *to_stream)
 		dump_title = TRUE;
 	if (getenv ("GDA_DATA_MODEL_NULL_AS_EMPTY")) /* Flawfinder: ignore */
 		null_as_empty = TRUE;
+	cstr = getenv ("GDA_DATA_MODEL_DUMP_TRUNCATE");
+	if (cstr) {
+		max_width = atoi (cstr);
+#ifdef TIOCGWINSZ
+		if (max_width < 0) {
+			struct winsize window_size;
+			if (ioctl (0,TIOCGWINSZ, &window_size) == 0)
+				max_width = (int) window_size.ws_col;
+		}
+#endif
+		if (max_width < 0)
+			max_width = 0;
+	}
 
-	str = real_gda_data_model_dump_as_string (model, FALSE, dump_rows, dump_title, null_as_empty, &error);
+	str = real_gda_data_model_dump_as_string (model, FALSE, dump_rows, dump_title, null_as_empty, max_width, &error);
 	if (str) {
 		g_fprintf (to_stream, "%s", str);
 		g_free (str);
 		if (dump_attrs) {
-			str = real_gda_data_model_dump_as_string (model, TRUE, dump_rows, dump_title, null_as_empty, &error);
+			str = real_gda_data_model_dump_as_string (model, TRUE, dump_rows, dump_title, null_as_empty, max_width, &error);
 			if (str) {
 				g_fprintf (to_stream, "%s", str);
 				g_free (str);
@@ -2092,7 +2144,8 @@ gda_data_model_dump (GdaDataModel *model, FILE *to_stream)
  * <itemizedlist>
  *   <listitem><para>GDA_DATA_MODEL_DUMP_ROW_NUMBERS: if set, the first column of the output will contain row numbers</para></listitem>
 *   <listitem><para>GDA_DATA_MODEL_DUMP_TITLE: if set, also dump the data model's title</para></listitem>
-*   <listitem><para>GDA_DATA_MODEL_DUMP_NULL_AS_EMPTY: if set, replace the 'NULL' string with an empty string for NULL values </para></listitem>
+*   <listitem><para>GDA_DATA_MODEL_NULL_AS_EMPTY: if set, replace the 'NULL' string with an empty string for NULL values </para></listitem>
+*   <listitem><para>GDA_DATA_MODEL_DUMP_TRUNCATE: if set to a numeric value, truncates the output to the width specified by the value. If the value is -1 then the actual terminal size (if it can be determined) is used</para></listitem>
  * </itemizedlist>
  * Returns: a new string.
  */
@@ -2102,6 +2155,8 @@ gda_data_model_dump_as_string (GdaDataModel *model)
 	gboolean dump_rows = FALSE;
 	gboolean dump_title = FALSE;
 	gboolean null_as_empty = FALSE;
+	gint max_width = 0; /* no truncate */
+	const gchar *cstr;
 
 	g_return_val_if_fail (GDA_IS_DATA_MODEL (model), NULL);
 
@@ -2111,8 +2166,20 @@ gda_data_model_dump_as_string (GdaDataModel *model)
 		dump_title = TRUE;
 	if (getenv ("GDA_DATA_MODEL_NULL_AS_EMPTY")) /* Flawfinder: ignore */
 		null_as_empty = TRUE;
-
-	return real_gda_data_model_dump_as_string (model, FALSE, dump_rows, dump_title, null_as_empty, NULL);
+	cstr = getenv ("GDA_DATA_MODEL_DUMP_TRUNCATE");
+	if (cstr) {
+		max_width = atoi (cstr);
+#ifdef TIOCGWINSZ
+		if (max_width < 0) {
+			struct winsize window_size;
+			if (ioctl (0,TIOCGWINSZ, &window_size) == 0)
+				max_width = (int) window_size.ws_col;
+		}
+#endif
+		if (max_width < 0)
+			max_width = 0;
+	}
+	return real_gda_data_model_dump_as_string (model, FALSE, dump_rows, dump_title, null_as_empty, max_width, NULL);
 }
 
 static void
@@ -2139,7 +2206,8 @@ string_get_dimensions (const gchar *string, gint *width, gint *rows)
 
 static gchar *
 real_gda_data_model_dump_as_string (GdaDataModel *model, gboolean dump_attributes, 
-				    gboolean dump_rows, gboolean dump_title, gboolean null_as_empty, GError **error)
+				    gboolean dump_rows, gboolean dump_title, gboolean null_as_empty,
+				    gint max_width, GError **error)
 {
 #define ERROR_STRING "####"
 #define MULTI_LINE_NO_SEPARATOR
@@ -2442,10 +2510,29 @@ real_gda_data_model_dump_as_string (GdaDataModel *model, gboolean dump_attribute
 			g_strfreev (cols_str [i]);
 		g_free (cols_str);
 	}
-	if (n_rows < 0)
-		g_string_append_printf (string, _("(error getting data)\n"));
+
+	/* status message */
+	g_string_append_c (string, '(');
+	if (n_rows > 0)
+		g_string_append_printf (string, ngettext("%d row", "%d rows", n_rows), n_rows);
 	else
-		g_string_append_printf (string, ngettext("(%d row)\n", "(%d rows)\n", n_rows), n_rows);
+		g_string_append_printf (string, _("0 row"));
+
+	GError **exceptions;
+	exceptions = gda_data_model_get_exceptions (model);
+	if (exceptions) {
+		gint i;
+		for (i = 0; exceptions[i]; i++) {
+			GError *ex;
+			ex = exceptions[i];
+			if (ex && ex->message) {
+				g_string_append (string, ", ");
+				g_string_append (string, ex->message);
+			}
+		}
+	}
+	g_string_append (string, ")\n");
+
 
  out:
 	if (ramodel)
@@ -2456,6 +2543,43 @@ real_gda_data_model_dump_as_string (GdaDataModel *model, gboolean dump_attribute
 	if (allok) {
 		str = string->str;
 		g_string_free (string, FALSE);
+		if (max_width > 0) {
+			/* truncate all lines */
+			GString *ns;
+			ns = g_string_sized_new (strlen (str));
+
+			gchar *ptr, *sptr;
+			gint len; /* number of characters since start of line */
+			for (ptr = str, sptr = ptr, len=0;
+			     *ptr;
+			     ptr = g_utf8_next_char (ptr), len++) {
+				if (*ptr == '\n') {
+					*ptr = 0;
+					len = 0;
+					g_string_append (ns, sptr);
+					g_string_append_c (ns, '\n');
+					*ptr = '\n';
+					sptr = ptr+1;
+				}
+				else if (len >= max_width) {
+					gchar c;
+					c = *ptr;
+					*ptr = 0;
+					len = 0;
+					g_string_append (ns, sptr);
+					g_string_append_c (ns, '\n');
+					*ptr = c;
+					for (; *ptr && (*ptr != '\n');
+					     ptr = g_utf8_next_char (ptr));
+					if (!*ptr)
+						break;
+					sptr = ptr+1;
+				}
+			}
+			g_free (str);
+			str = ns->str;
+			g_string_free (ns, FALSE);
+		}
 	}
 	else {
 		str = NULL;
