@@ -1,5 +1,5 @@
-/* GDA library
- * Copyright (C) 2009 - 2010 The GNOME Foundation.
+/*
+ * Copyright (C) 2009 - 2011 The GNOME Foundation.
  *
  * AUTHORS:
  *      Vivien Malerba <malerba@gnome-db.org>
@@ -166,7 +166,7 @@ gda_tree_node_class_init (GdaTreeNodeClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GdaTreeNodeClass, node_deleted),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__BOXED,
+			      g_cclosure_marshal_VOID__STRING,
 			      G_TYPE_NONE, 1, G_TYPE_STRING);
 
 	klass->node_changed = m_node_changed;
@@ -261,7 +261,7 @@ m_node_deleted (GdaTreeNode *reporting, const gchar *relative_path)
 }
 
 static void
-gda_tree_node_init (GdaTreeNode *tnode, GdaTreeNodeClass *klass)
+gda_tree_node_init (GdaTreeNode *tnode, G_GNUC_UNUSED GdaTreeNodeClass *klass)
 {
 	g_return_if_fail (GDA_IS_TREE_NODE (tnode));
 
@@ -322,7 +322,8 @@ gda_tree_node_get_type (void)
                         NULL,
                         sizeof (GdaTreeNode),
                         0,
-                        (GInstanceInitFunc) gda_tree_node_init
+                        (GInstanceInitFunc) gda_tree_node_init,
+			0
                 };
 
                 g_static_mutex_lock (&registering);
@@ -347,6 +348,9 @@ gda_tree_node_set_property (GObject *object,
 		case PROP_NAME:
 			gda_attributes_manager_set (gda_tree_node_attributes_manager, tnode, GDA_ATTRIBUTE_NAME, value);
 			break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
 		}	
 	}
 }
@@ -368,6 +372,9 @@ gda_tree_node_get_property (GObject *object,
 			g_value_set_string (value, cvalue ? g_value_get_string (cvalue): NULL);
 			break;
 		}
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
 		}
 	}	
 }
@@ -717,10 +724,11 @@ gda_tree_node_get_node_attribute (GdaTreeNode *node, const gchar *attribute)
  * gda_tree_node_set_node_attribute:
  * @node: a #GdaTreeNode
  * @attribute: attribute name
- * @value: a #GValue, or %NULL
+ * @value: (transfer none): a #GValue, or %NULL
  * @destroy: a function to be called when @attribute is not needed anymore, or %NULL
  *
- * Set the value associated to a named attribute. The @attribute string is 'stolen' by this method, and
+ * Set the value associated to a named attribute. The @attribute string is used AS IT IS by this method (eg.
+ * no copy of it is made), and
  * the memory it uses will be freed using the @destroy function when no longer needed (if @destroy is %NULL,
  * then the string will not be freed at all).
  *
@@ -730,8 +738,8 @@ gda_tree_node_get_node_attribute (GdaTreeNode *node, const gchar *attribute)
  * For example one would use it as:
  *
  * <code>
- * gda_tree_node_set_node_attribute (node, g_strdup (my_attribute), g_free, my_value);
- * gda_tree_node_set_node_attribute (node, GDA_ATTRIBUTE_NAME, NULL, my_value);
+ * gda_tree_node_set_node_attribute (node, g_strdup (my_attribute), my_value, g_free);
+ * gda_tree_node_set_node_attribute (node, GDA_ATTRIBUTE_NAME, my_value, NULL);
  * </code>
  *
  * If there is already an attribute named @attribute set, then its value is replaced with the new value (@value is
