@@ -1,6 +1,5 @@
 /* 
- * GDA common library
- * Copyright (C) 2007 - 2010 The GNOME Foundation.
+ * Copyright (C) 2007 - 2011 The GNOME Foundation.
  *
  * AUTHORS:
  *      Vivien Malerba <malerba@gnome-db.org>
@@ -28,6 +27,7 @@
 #include "gda-vconnection-data-model-private.h"
 #include <sqlite3.h>
 #include <libgda/gda-connection-private.h>
+#include <libgda/gda-connection-internal.h>
 #include <libgda/gda-data-model-iter.h>
 #include <libgda/gda-blob-op.h>
 #include "../gda-sqlite.h"
@@ -52,6 +52,12 @@ static gboolean       gda_vprovider_data_model_open_connection (GdaServerProvide
 								gpointer cb_data);
 static gboolean       gda_vprovider_data_model_close_connection (GdaServerProvider *provider,
 								 GdaConnection *cnc);
+static GObject        *gda_vprovider_data_model_statement_execute (GdaServerProvider *provider, GdaConnection *cnc,
+								   GdaStatement *stmt, GdaSet *params,
+								   GdaStatementModelUsage model_usage,
+								   GType *col_types, GdaSet **last_inserted_row,
+								   guint *task_id, GdaServerProviderExecCallback async_cb,
+								   gpointer cb_data, GError **error);
 static const gchar   *gda_vprovider_data_model_get_name (GdaServerProvider *provider);
 
 /*
@@ -69,6 +75,7 @@ gda_vprovider_data_model_class_init (GdaVproviderDataModelClass *klass)
 	server_class->create_connection = gda_vprovider_data_model_create_connection;
 	server_class->open_connection = gda_vprovider_data_model_open_connection;
 	server_class->close_connection = gda_vprovider_data_model_close_connection;
+	server_class->statement_execute = gda_vprovider_data_model_statement_execute;
 
 	server_class->get_name = gda_vprovider_data_model_get_name;
 
@@ -147,6 +154,54 @@ gda_vprovider_data_model_new (void)
         return provider;
 }
 
+/*
+ * Note about RowIDs and how SQLite uses them:
+ *
+ * SQLite considers that each virtual table has unique row IDs, absolute value available all the time.
+ * When an UPDATE or DELETE statement is executed, SQLite does the following:
+ *  - xBegin (table): request a transaction start
+ *  - xOpen (table): create a new cursor
+ *  - xFilter (cursor): initialize the cursor
+ *  - moves the cursor one step at a time up to the end (xEof, xColumn and xNext). If it finds a
+ *    row needing to be updated or deleted, it calls xRowid (cursor) to get the RowID
+ *  - xClose (cursor): free the now useless cursor
+ *  - calls xUpdate (table) as many times as needed (one time for each row where xRowid was called)
+ *  - xSync (table)
+ *  - xCommit (table)
+ *
+ * This does not work well with Libgda because it does not know how to define a unique RowID for
+ * a table: it defines the RowID as being the position of the row in the data model, which changes
+ * each time the data model used for the virtuel table changes.
+ *
+ * Moreover, when the cursor has reached the end of the data model, it may not be possible to move
+ * it backwards and so the data at any given row is not accessible anymore. The solution to this
+ * problem is, each time the xRowid() is called, to copy the row in memory, using the table->rowid_hash
+ * hash table.
+ */
+
+#ifdef GDA_DEBUG_VIRTUAL
+#define TRACE(table,cursor) g_print ("== %s (table=>%p cursor=>%p)\n", __FUNCTION__, (table), (cursor))
+#else
+#define TRACE(table,cursor)
+#endif
+
+typedef struct {
+	sqlite3_vtab                 base;
+	GdaVconnectionDataModel     *cnc;
+	GdaVConnectionTableData     *td;
+
+	GdaDataModel                *rowid_hash_model; /* data model used to build the rowid_hash's
+							* contents. No ref held there as it's never
+							* dereferenced */
+	GHashTable                  *rowid_hash; /* key = a gint64 rowId, value = a GdaRow */
+} VirtualTable;
+
+typedef struct {
+	sqlite3_vtab_cursor      base; /* base.pVtab is a pointer to the sqlite3_vtab virtual table */
+	GdaDataModel            *model; /* in which @iter iterates */
+	GdaDataModelIter        *iter;
+} VirtualCursor;
+
 /* module creation */
 static int virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlite3_vtab **ppVtab, char **pzErr);
 static int virtualConnect (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlite3_vtab **ppVtab, char **pzErr);
@@ -266,31 +321,79 @@ gda_vprovider_data_model_close_connection (GdaServerProvider *provider, GdaConne
 	return GDA_SERVER_PROVIDER_CLASS (parent_class)->close_connection (GDA_SERVER_PROVIDER (provider), cnc);
 }
 
+static GObject *
+gda_vprovider_data_model_statement_execute (GdaServerProvider *provider, GdaConnection *cnc,
+					    GdaStatement *stmt, GdaSet *params,
+					    GdaStatementModelUsage model_usage,
+					    GType *col_types, GdaSet **last_inserted_row,
+					    guint *task_id, GdaServerProviderExecCallback async_cb,
+					    gpointer cb_data, GError **error)
+{
+	GObject *retval;
+	if (async_cb) {
+                g_set_error (error, GDA_SERVER_PROVIDER_ERROR, GDA_SERVER_PROVIDER_METHOD_NON_IMPLEMENTED_ERROR,
+                             "%s", _("Provider does not support asynchronous statement execution"));
+                return NULL;
+        }
+        
+	retval = GDA_SERVER_PROVIDER_CLASS (parent_class)->statement_execute (provider, cnc, stmt, params,
+									      model_usage, col_types,
+									      last_inserted_row, task_id,
+									      async_cb, cb_data, error);
+	if (retval) {
+		gchar *sql;
+		sql = gda_statement_to_sql (stmt, params, NULL);
+		if (sql) {
+			gchar *ptr = NULL;
+			/* look for DROP TABLE to signal table removal */
+			if (! g_ascii_strncasecmp (sql, "DROP", 4))
+				ptr = sql + 4;
+			else if (! g_ascii_strncasecmp (sql, "CREATE", 6))
+				ptr = sql + 6;
+			if (ptr) {
+				/* skip spaces */
+				for (; *ptr && (g_ascii_isspace (*ptr) || (*ptr == '\n')); ptr++);
+
+				if (! g_ascii_strncasecmp (ptr, "TABLE", 5)) {
+					/* skip spaces */
+					gchar delim = 0;
+					gchar *table_name, *quoted;
+					for (ptr = ptr+5;
+					     *ptr && (g_ascii_isspace (*ptr) || (*ptr == '\n'));
+					     ptr++);
+					if (*ptr == '\'') {
+						delim = '\'';
+						ptr++;
+					}
+					else if (*ptr == '"') {
+						delim = '"';
+						ptr++;
+					}
+					table_name = ptr;
+					if (delim)
+						for (; *ptr && (*ptr != delim) ; ptr++);
+					else
+						for (; *ptr && g_ascii_isalnum (*ptr); ptr++);
+					*ptr = 0;
+					quoted = _gda_connection_compute_table_virtual_name (GDA_CONNECTION (cnc), table_name);
+					/*g_print ("%s() emits the 'vtable-dropped' signal for table [%s]\n",
+					  __FUNCTION__, quoted);*/
+					g_signal_emit_by_name (cnc, "vtable-dropped", quoted);
+					g_free (quoted);
+				}
+			}
+			g_free (sql);
+		}
+	}
+	return retval;
+}
+
 static const gchar *
 gda_vprovider_data_model_get_name (G_GNUC_UNUSED GdaServerProvider *provider)
 {
 	return "Virtual";
 }
 
-/* module implementation */
-#ifdef GDA_DEBUG_VIRTUAL
-  #define TRACE(table) g_print ("== %s (table=>%p)\n", __FUNCTION__, (table))
-#else
-  #define TRACE(table)
-#endif
-
-typedef struct {
-	sqlite3_vtab                 base;
-	GdaVconnectionDataModel     *cnc;
-	GdaVConnectionTableData     *td;
-} VirtualTable;
-
-typedef struct {
-	sqlite3_vtab_cursor      base; /* base.pVtab is a pointer to the sqlite3_vtab virtual table */
-	GdaDataModelIter        *iter;
-	gint                     ncols;
-} VirtualCursor;
-
 static int
 virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlite3_vtab **ppVtab, char **pzErr)
 {
@@ -300,7 +403,7 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 	gchar *spec_name, *tmp;
 	GdaVConnectionTableData *td;
 
-	TRACE (NULL);
+	TRACE (NULL, NULL);
 
 	/* find Spec */
 	g_assert (argc == 4);
@@ -325,22 +428,22 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 		td->real_model = td->spec->data_model;
 		g_object_ref (td->real_model);
 	}
-	else  {
-		GError *error = NULL;
-		if (!td->columns)
-			td->columns = td->spec->create_columns_func (td->spec, &error);
-		if (! td->columns) {
-			if (error && error->message) {
-				int len = strlen (error->message) + 1;
-				*pzErr = SQLITE3_CALL (sqlite3_malloc) (sizeof (gchar) * len);
-				memcpy (*pzErr, error->message, len); /* Flawfinder: ignore */
-			}
-			else 
-				*pzErr = SQLITE3_CALL (sqlite3_mprintf) (_("Could not compute virtual table's columns"));
-			return SQLITE_ERROR;
+
+	GError *error = NULL;
+	if (!td->columns && td->spec->create_columns_func)
+		td->columns = td->spec->create_columns_func (td->spec, &error);
+	if (!td->columns) {
+		if (error && error->message) {
+			int len = strlen (error->message) + 1;
+			*pzErr = SQLITE3_CALL (sqlite3_malloc) (sizeof (gchar) * len);
+			memcpy (*pzErr, error->message, len); /* Flawfinder: ignore */
 		}
-		ncols = g_list_length (td->columns);
+		else 
+			*pzErr = SQLITE3_CALL (sqlite3_mprintf) (_("Could not compute virtual table's columns"));
+		return SQLITE_ERROR;
 	}
+	ncols = g_list_length (td->columns);
+	td->n_columns = ncols;
 
 	/* create the CREATE TABLE statement */
 	sql = g_string_new ("CREATE TABLE ");
@@ -356,10 +459,7 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 
 		if (i != 0)
 			g_string_append (sql, ", ");
-		if (td->columns)
-			column = g_list_nth_data (td->columns, i);
-		else
-			column = gda_data_model_describe_column (td->spec->data_model, i);
+		column = g_list_nth_data (td->columns, i);
 		if (!column) {
 			*pzErr = SQLITE3_CALL (sqlite3_mprintf) (_("Can't get data model description for column %d"), i);
 			g_string_free (sql, TRUE);
@@ -375,12 +475,14 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 		gtype = gda_column_get_g_type (column);
 		type = g_type_name (gtype);
 		if (!type) {
-			*pzErr = SQLITE3_CALL (sqlite3_mprintf) (_("Can't get data model's column type or type for column %d"), i);
+			*pzErr = SQLITE3_CALL (sqlite3_mprintf) (_("Can't get data model's column type for column %d"), i);
 			g_string_free (sql, TRUE);
 			return SQLITE_ERROR;
 		}
-		else if ((gtype == GDA_TYPE_BLOB) || (gtype == GDA_TYPE_BINARY))
+		else if (gtype == GDA_TYPE_BLOB)
 			type = "blob";
+		else if (gtype == GDA_TYPE_BINARY)
+			type = "binary";
 		else if (gtype == G_TYPE_STRING)
 			type = "string";
 		else if ((gtype == G_TYPE_INT) || (gtype == G_TYPE_UINT) || 
@@ -429,7 +531,7 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 		return SQLITE_ERROR;
 	}
 
-	/*g_print ("VIRTUAL TABLE: %s\n", sql->str);*/
+	/*g_print ("VIRTUAL TABLE [%p]: %s\n", vtable, sql->str);*/
 	g_string_free (sql, TRUE);
 
 	return SQLITE_OK;
@@ -438,7 +540,7 @@ virtualCreate (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlit
 static int
 virtualConnect (sqlite3 *db, void *pAux, int argc, const char *const *argv, sqlite3_vtab **ppVtab, char **pzErr)
 {
-	TRACE (NULL);
+	TRACE (NULL, NULL);
 
 	return virtualCreate (db, pAux, argc, argv, ppVtab, pzErr);
 }
@@ -448,8 +550,10 @@ virtualDisconnect (sqlite3_vtab *pVtab)
 {
 	VirtualTable *vtable = (VirtualTable *) pVtab;
 
-	TRACE (pVtab);
+	TRACE (pVtab, NULL);
 
+	if (vtable->rowid_hash)
+		g_hash_table_destroy (vtable->rowid_hash);
 	g_free (vtable);
 	return SQLITE_OK;
 }
@@ -457,7 +561,7 @@ virtualDisconnect (sqlite3_vtab *pVtab)
 static int
 virtualDestroy (sqlite3_vtab *pVtab)
 {
-	TRACE (pVtab);
+	TRACE (pVtab, NULL);
 
 	return virtualDisconnect (pVtab);
 }
@@ -467,11 +571,10 @@ virtualOpen (sqlite3_vtab *pVTab, sqlite3_vtab_cursor **ppCursor)
 {
 	VirtualCursor *cursor;
 
-	TRACE (pVTab);
-
 	/* create empty cursor */
 	cursor = g_new0 (VirtualCursor, 1);
 	*ppCursor = (sqlite3_vtab_cursor*) cursor;
+	TRACE (pVTab, cursor);
 
 	return SQLITE_OK;
 }
@@ -481,12 +584,15 @@ virtualClose (sqlite3_vtab_cursor *cur)
 {
 	VirtualCursor *cursor = (VirtualCursor*) cur;
 
-	TRACE (cur->pVtab);
+	TRACE (cur->pVtab, cur);
 
 	if (cursor->iter)
 		g_object_unref (cursor->iter);
-	/* FIXME: destroy table->spec->model and table->wrapper */
+	if (cursor->model)
+		g_object_unref (cursor->model);
+
 	g_free (cur);
+
 	return SQLITE_OK;
 }
 
@@ -495,7 +601,7 @@ virtualEof (sqlite3_vtab_cursor *cur)
 {
 	VirtualCursor *cursor = (VirtualCursor*) cur;
 
-	TRACE (cur->pVtab);
+	TRACE (cur->pVtab, cur);
 
 	if (gda_data_model_iter_is_valid (cursor->iter))
 		return FALSE;
@@ -509,7 +615,10 @@ virtualNext (sqlite3_vtab_cursor *cur)
 	VirtualCursor *cursor = (VirtualCursor*) cur;
 	/*VirtualTable *vtable = (VirtualTable*) cur->pVtab;*/
 
-	TRACE (cur->pVtab);
+	TRACE (cur->pVtab, cur);
+
+	if (gda_data_model_get_exceptions (cursor->model))
+		return SQLITE_IOERR_TRUNCATE;
 
 	if (!gda_data_model_iter_move_next (cursor->iter)) {
 		if (gda_data_model_iter_is_valid (cursor->iter))
@@ -525,11 +634,11 @@ virtualColumn (sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int i)
 {
 	VirtualCursor *cursor = (VirtualCursor*) cur;
 
-	TRACE (cur->pVtab);
+	TRACE (cur->pVtab, cur);
 
 	GdaHolder *param;
 	
-	if (i == cursor->ncols) {
+	if (i == ((VirtualTable*) cur->pVtab)->td->n_columns) {
 		/* private hidden column, which returns the row number */
 		SQLITE3_CALL (sqlite3_result_int) (ctx, gda_data_model_iter_get_row (cursor->iter));
 		return SQLITE_OK;
@@ -580,10 +689,45 @@ static int
 virtualRowid (sqlite3_vtab_cursor *cur, sqlite_int64 *pRowid)
 {
 	VirtualCursor *cursor = (VirtualCursor*) cur;
+	VirtualTable *vtable = (VirtualTable*) cur->pVtab;
 
-	TRACE (cur->pVtab);
+	TRACE (vtable, cur);
 
 	*pRowid = gda_data_model_iter_get_row (cursor->iter);
+	if (! vtable->rowid_hash || (vtable->rowid_hash_model == vtable->td->real_model)) {
+		if (! vtable->rowid_hash) {
+			vtable->rowid_hash = g_hash_table_new_full (g_int64_hash, g_int64_equal,
+								    g_free,
+								    (GDestroyNotify) g_object_unref);
+			vtable->rowid_hash_model = vtable->td->real_model;
+		}
+
+		GdaRow *grow;
+		gint i, ncols;
+		gint64 *hid;
+		ncols = g_list_length (vtable->td->columns);
+		
+		grow = gda_row_new (ncols);
+		for (i = 0; i < ncols; i++) {
+			const GValue *cvalue;
+			GValue *gvalue;
+			cvalue = gda_data_model_iter_get_value_at (cursor->iter, i);
+			gvalue = gda_row_get_value (grow, i);
+			if (cvalue) {
+				if (G_VALUE_TYPE (cvalue) != GDA_TYPE_NULL) {
+					g_value_init (gvalue, G_VALUE_TYPE (cvalue));
+					g_value_copy (cvalue, gvalue);
+				}
+ 			}
+			else
+				gda_row_invalidate_value (grow, gvalue);
+		}
+		
+		hid = g_new (gint64, 1);
+		*hid = *pRowid;
+		g_hash_table_insert (vtable->rowid_hash, hid, grow);
+	}
+
 	return SQLITE_OK;
 }
 
@@ -705,23 +849,20 @@ virtualFilter (sqlite3_vtab_cursor *pVtabCursor, int idxNum, const char *idxStr,
 	VirtualCursor *cursor = (VirtualCursor*) pVtabCursor;
 	VirtualTable *vtable = (VirtualTable*) pVtabCursor->pVtab;
 
-	TRACE (pVtabCursor->pVtab);
+	TRACE (pVtabCursor->pVtab, pVtabCursor);
 
 	virtual_table_manage_real_data_model (vtable, idxNum, idxStr, argc, argv);
 	if (! vtable->td->real_model)
 		return SQLITE_ERROR;
 
 	/* initialize cursor */
-	if (GDA_IS_DATA_PROXY (vtable->td->real_model)) {
+	if (GDA_IS_DATA_PROXY (vtable->td->real_model))
 		cursor->iter = g_object_new (GDA_TYPE_DATA_MODEL_ITER,
 					     "data-model", vtable->td->real_model, NULL);
-		cursor->ncols = gda_data_model_get_n_columns (GDA_DATA_MODEL (vtable->td->real_model));
-	}
-	else {
+	else
 		cursor->iter = gda_data_model_create_iter (vtable->td->real_model);
-		cursor->ncols = gda_data_model_get_n_columns (GDA_DATA_MODEL (vtable->td->real_model));
-	}
 
+	cursor->model = g_object_ref (vtable->td->real_model);
 	gda_data_model_iter_move_next (cursor->iter);
 	return SQLITE_OK;
 }
@@ -770,43 +911,48 @@ index_info_dump (sqlite3_index_info *pIdxInfo, gboolean dump_out)
 static void
 map_sqlite3_info_to_gda_filter (sqlite3_index_info *info, GdaVconnectionDataModelFilter *filter)
 {
-	gint i;
 	memset (filter, 0, sizeof (GdaVconnectionDataModelFilter));
-	filter->nConstraint = info->nConstraint;
 	if (info->nConstraint > 0) {
-		filter->aConstraint = g_new (struct GdaVirtualConstraint, filter->nConstraint);
-		filter->aConstraintUsage = g_new (struct GdaVirtualConstraintUsage, filter->nConstraint);
-		for (i = 0; i < info->nConstraint; i++) {
-			filter->aConstraint[i].iColumn = info->aConstraint[i].iColumn;
+		gint i, j;
+		filter->aConstraint = g_new (struct GdaVirtualConstraint, info->nConstraint);
+		filter->aConstraintUsage = g_new (struct GdaVirtualConstraintUsage, info->nConstraint);
+		for (i = 0, j = 0; i < info->nConstraint; i++) {
+			if (! info->aConstraint[i].usable)
+				continue;
+
+			filter->aConstraint[j].iColumn = info->aConstraint[i].iColumn;
 			switch (info->aConstraint[i].op) {
 			case SQLITE_INDEX_CONSTRAINT_EQ:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_EQ;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_EQ;
 				break;
 			case SQLITE_INDEX_CONSTRAINT_GT:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_GT;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_GT;
 				break;
 			case SQLITE_INDEX_CONSTRAINT_LE:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_LEQ;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_LEQ;
 				break;
 			case SQLITE_INDEX_CONSTRAINT_LT:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_LT;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_LT;
 				break;
 			case SQLITE_INDEX_CONSTRAINT_GE:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_GEQ;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_GEQ;
 				break;
 			case SQLITE_INDEX_CONSTRAINT_MATCH:
-				filter->aConstraint[i].op = GDA_SQL_OPERATOR_TYPE_REGEXP;
+				filter->aConstraint[j].op = GDA_SQL_OPERATOR_TYPE_REGEXP;
 				break;
 			default:
 				g_assert_not_reached ();
 			}
 
-			filter->aConstraintUsage[i].argvIndex = 0;
-			filter->aConstraintUsage[i].omit = FALSE;
+			filter->aConstraintUsage[j].argvIndex = 0;
+			filter->aConstraintUsage[j].omit = FALSE;
+			j++;
 		}
+		filter->nConstraint = j;
 	}
 	filter->nOrderBy = info->nOrderBy;
 	if (filter->nOrderBy > 0) {
+		gint i;
 		filter->aOrderBy = g_new (struct GdaVirtualOrderby, filter->nOrderBy);
 		for (i = 0; i < info->nOrderBy; i++) {
 			filter->aOrderBy[i].iColumn = info->aOrderBy[i].iColumn;
@@ -825,12 +971,15 @@ map_sqlite3_info_to_gda_filter (sqlite3_index_info *info, GdaVconnectionDataMode
 static void
 map_consume_gda_filter_to_sqlite3_info (GdaVconnectionDataModelFilter *filter, sqlite3_index_info *info)
 {
-	gint i;
 	g_assert (filter->nConstraint == info->nConstraint);
 	if (info->nConstraint > 0) {
-		for (i = 0; i < info->nConstraint; i++) {
-			info->aConstraintUsage[i].argvIndex = filter->aConstraintUsage[i].argvIndex;
-			info->aConstraintUsage[i].omit = filter->aConstraintUsage[i].omit ? 1 : 0;
+		gint i, j;
+		for (i = 0, j = 0; i < info->nConstraint; i++) {
+			if (! info->aConstraint[i].usable)
+				continue;
+			info->aConstraintUsage[i].argvIndex = filter->aConstraintUsage[j].argvIndex;
+			info->aConstraintUsage[i].omit = filter->aConstraintUsage[j].omit ? 1 : 0;
+			j++;
 		}
 		g_free (filter->aConstraint);
 		g_free (filter->aConstraintUsage);
@@ -850,7 +999,7 @@ virtualBestIndex (sqlite3_vtab *tab, sqlite3_index_info *pIdxInfo)
 {
 	VirtualTable *vtable = (VirtualTable *) tab;
 
-	TRACE (tab);
+	TRACE (tab, NULL);
 #ifdef GDA_DEBUG_VIRTUAL
 	index_info_dump (pIdxInfo, FALSE);
 #endif
@@ -869,6 +1018,22 @@ virtualBestIndex (sqlite3_vtab *tab, sqlite3_index_info *pIdxInfo)
 	return SQLITE_OK;
 }
 
+/*
+ *
+ * Returns: >= 0 if Ok, -1 on error
+ */
+static gint
+param_name_to_number (gint maxrows, const gchar *str)
+{
+	gchar *endptr [1];
+	long int i;
+	i = strtol (str, endptr, 10);
+	if ((**endptr == '\0') && (i < maxrows) && (i >= 0))
+		return i;
+	else
+		return -1;
+}
+
 /*
  *    apData[0]  apData[1]  apData[2..]
  *
@@ -885,35 +1050,234 @@ virtualUpdate (sqlite3_vtab *tab, int nData, sqlite3_value **apData, sqlite_int6
 {
 	VirtualTable *vtable = (VirtualTable *) tab;
 	const gchar *api_misuse_error = NULL;
+	gint optype; /* 1 => DELETE
+		      * 2 => INSERT
+		      * 3 => UPDATE
+		      */
 
-	TRACE (tab);
+	TRACE (tab, NULL);
+
+	/* determine operation type */
+	if (nData == 1)
+		optype = 1;
+	else if ((nData > 1) && (SQLITE3_CALL (sqlite3_value_type) (apData[0]) == SQLITE_NULL)) {
+		optype = 2;
+		if (SQLITE3_CALL (sqlite3_value_type) (apData[1]) != SQLITE_NULL) {
+			/* argc>1 and argv[0] is not NULL: rowid is imposed by SQLite
+			 * which is not supported */
+			return SQLITE_READONLY;
+		}
+	}
+	else if ((nData > 1) && (SQLITE3_CALL (sqlite3_value_type) (apData[0]) == SQLITE_INTEGER)) {
+		optype = 3;
+		if (SQLITE3_CALL (sqlite3_value_int) (apData[0]) != 
+		    SQLITE3_CALL (sqlite3_value_int) (apData[1])) {
+			/* argc>1 and argv[0]==argv[1]: rowid is imposed by SQLite 
+			 * which is not supported */
+			return SQLITE_READONLY;
+		}
+	}
+	else
+		return SQLITE_MISUSE;
+
+	/* handle data model */
+	if (! vtable->td->real_model) {
+		virtual_table_manage_real_data_model (vtable, -1, NULL, 0, NULL);
+		if (! vtable->td->real_model)
+			return SQLITE_ERROR;
+	}
+
+	GdaDataModelAccessFlags access_flags;
+	access_flags = gda_data_model_get_access_flags (vtable->td->real_model);
+	if (((optype == 1) && ! (access_flags & GDA_DATA_MODEL_ACCESS_DELETE)) ||
+	    ((optype == 2) && ! (access_flags & GDA_DATA_MODEL_ACCESS_INSERT)) ||
+	    ((optype == 3) && ! (access_flags & GDA_DATA_MODEL_ACCESS_UPDATE))) {
+		/* we can't use vtable->td->real_model because it can't be accessed correctly */
+		if (! GDA_IS_DATA_SELECT (vtable->td->real_model)) {
+			tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+				(_("Data model representing the table is read only"));
+			return SQLITE_READONLY;
+		}
+		
+		/* determine parameters required to execute MOD statement */
+		GdaStatement *stmt = NULL;
+		ParamType ptype;
+		switch (optype) {
+		case 1:
+			ptype = PARAMS_DELETE;
+			if (! vtable->td->modif_stmt [ptype])
+				g_object_get (vtable->td->real_model, "delete-stmt", &stmt, NULL);
+			break;
+		case 2:
+			ptype = PARAMS_INSERT;
+			if (! vtable->td->modif_stmt [ptype])
+				g_object_get (vtable->td->real_model, "insert-stmt", &stmt, NULL);
+			break;
+		case 3:
+			ptype = PARAMS_UPDATE;
+			if (! vtable->td->modif_stmt [ptype])
+				g_object_get (vtable->td->real_model, "update-stmt", &stmt, NULL);
+			break;
+		default:
+			g_assert_not_reached ();
+		}
+		
+		if (! vtable->td->modif_stmt [ptype]) {
+			if (! stmt) {
+				tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+					(_("No statement provided to modify the data model "
+					   "representing the table"));
+				return SQLITE_READONLY;
+			}
+		
+			GdaSet *params;
+			if (! gda_statement_get_parameters (stmt, &params, NULL) || !params) {
+				tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+					(_("Invalid statement provided to modify the data model "
+					   "representing the table"));
+				g_object_unref (stmt);
+				return SQLITE_READONLY;
+			}
+			vtable->td->modif_stmt [ptype] = stmt;
+			vtable->td->modif_params [ptype] = params;
+		}
+		stmt = vtable->td->modif_stmt [ptype];
+		
+		/* bind parameters */
+		GSList *list;
+		for (list = vtable->td->modif_params [ptype]->holders; list; list = list->next) {
+			const gchar *id;
+			GdaHolder *holder = GDA_HOLDER (list->data);
+			gboolean holder_value_set = FALSE;
+			
+			id = gda_holder_get_id (holder);
+			if (!id) {
+				tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+					(_("Invalid parameter in statement provided to modify "
+					   "the data model representing the table"));
+				return SQLITE_READONLY;
+			}
+			if (*id == '+' && id[1]) {
+				long int i;
+				i = param_name_to_number (vtable->td->n_columns, id+1);
+				if (i >= 0) {
+					GType type;
+					GValue *value;
+					type = gda_column_get_g_type (gda_data_model_describe_column (vtable->td->real_model, i));
+					if ((type != GDA_TYPE_NULL) && SQLITE3_CALL (sqlite3_value_text) (apData [i+2]))
+						value = gda_value_new_from_string ((const gchar*) SQLITE3_CALL (sqlite3_value_text) (apData [i+2]), type);
+					else
+						value = gda_value_new_null ();
+					if (gda_holder_take_value (holder, value, NULL))
+						holder_value_set = TRUE;
+				}
+			}
+			else if (*id == '-') {
+				if (! vtable->rowid_hash) {
+					tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+						(_("Could not retreive row to delete"));
+					return SQLITE_READONLY;
+				}
+
+				gint64 rowid = SQLITE3_CALL (sqlite3_value_int64) (apData [0]);
+				GdaRow *grow = NULL;
+				if (vtable->rowid_hash_model == vtable->td->real_model)
+					grow = g_hash_table_lookup (vtable->rowid_hash, &rowid);
+				if (!grow) {
+					tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+						(_("Could not retreive row to delete"));
+					return SQLITE_READONLY;
+				}
+
+				long int i;
+				i = param_name_to_number (vtable->td->n_columns, id+1);
+				if (i >= 0) {
+					GValue *value;
+					value = gda_row_get_value (grow, i);
+					if (gda_holder_set_value (holder, value, NULL))
+						holder_value_set = TRUE;
+				}
+			}
+			
+			if (! holder_value_set) {
+				GdaSet *exec_set;
+				GdaHolder *eh;
+				g_object_get (vtable->td->real_model, "exec-params",
+					      &exec_set, NULL);
+				if (! exec_set) {
+					/* can't give value to param named @id */
+					tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+						(_("Invalid parameter in statement provided to modify "
+						   "the data model representing the table"));
+					return SQLITE_READONLY;
+				}
+				eh = gda_set_get_holder (exec_set, id);
+				if (! eh ||
+				    ! gda_holder_set_bind (holder, eh, NULL)) {
+					/* can't give value to param named @id */
+					tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+						(_("Invalid parameter in statement provided to modify "
+						   "the data model representing the table"));
+					return SQLITE_READONLY;
+				}
+			}
+		}
+
+		GdaConnection *cnc;
+		cnc = gda_data_select_get_connection (GDA_DATA_SELECT (vtable->td->real_model));
+		
+#ifdef GDA_DEBUG_NO
+		gchar *sql;
+		GError *lerror = NULL;
+		sql = gda_statement_to_sql (stmt, NULL, NULL);
+		g_print ("SQL: [%s] ", sql);
+		g_free (sql);
+		sql = gda_statement_to_sql_extended (stmt, cnc, params, GDA_STATEMENT_SQL_PRETTY, NULL, &lerror);
+		if (sql) {
+			g_print ("With params: [%s]\n", sql);
+			g_free (sql);
+		}
+		else {
+			g_print ("params ERROR [%s]\n", lerror && lerror->message ? lerror->message : "No detail");
+		}
+		g_clear_error (&lerror);
+#endif
+		
+		if (!cnc ||
+		    (gda_connection_statement_execute_non_select (cnc, stmt,
+								  vtable->td->modif_params [ptype],
+								  NULL, NULL) == -1)) {
+			/* failed to execute */
+			tab->zErrMsg = SQLITE3_CALL (sqlite3_mprintf)
+				(_("Failed to execute the statement provided to modify "
+				   "the data model representing the table"));
+			return SQLITE_READONLY;
+		}
+		return SQLITE_OK;
+	}
 
 	/* REM: when using the values of apData[], the limit is
 	 * (nData -1 ) and not nData because the last column of the corresponding CREATE TABLE ...
 	 * is an internal hidden field which does not correspond to any column of the real data model
 	 */
 
-	if (nData == 1) {
+	if (optype == 1) {
 		/* DELETE */
 		if (SQLITE3_CALL (sqlite3_value_type) (apData[0]) == SQLITE_INTEGER) {
 			gint rowid = SQLITE3_CALL (sqlite3_value_int) (apData [0]);
-			return gda_data_model_remove_row (vtable->td->real_model, rowid, NULL) ? SQLITE_OK : SQLITE_READONLY;
+			return gda_data_model_remove_row (vtable->td->real_model, rowid, NULL) ?
+				SQLITE_OK : SQLITE_READONLY;
 		}
 		else {
 			api_misuse_error = "argc==1 and argv[0] is not an integer";
 			goto api_misuse;
 		}
 	}
-	else if ((nData > 1) && (SQLITE3_CALL (sqlite3_value_type) (apData[0]) == SQLITE_NULL)) {
+	else if (optype == 2) {
 		/* INSERT */
 		gint newrow, i;
 		GList *values = NULL;
 		
-		if (SQLITE3_CALL (sqlite3_value_type) (apData[1]) != SQLITE_NULL) {
-			/* argc>1 and argv[0] is not NULL: rowid is imposed by SQLite which is not supported */
-			return SQLITE_READONLY;
-		}
-
 		for (i = 2; i < (nData - 1); i++) {
 			GType type;
 			GValue *value;
@@ -922,7 +1286,8 @@ virtualUpdate (sqlite3_vtab *tab, int nData, sqlite3_value **apData, sqlite_int6
 				value = gda_value_new_from_string ((const gchar*) SQLITE3_CALL (sqlite3_value_text) (apData [i]), type);
 			else
 				value = gda_value_new_null ();
-			/*g_print ("TXT #%s# => value %p (type=%s) apData[]=%p\n", SQLITE3_CALL (sqlite3_value_text) (apData [i]), value,
+			/*g_print ("TXT #%s# => value %p (type=%s) apData[]=%p\n",
+			  SQLITE3_CALL (sqlite3_value_text) (apData [i]), value,
 			  g_type_name (type), apData[i]);*/
 			values = g_list_prepend (values, value);
 		}
@@ -936,14 +1301,11 @@ virtualUpdate (sqlite3_vtab *tab, int nData, sqlite3_value **apData, sqlite_int6
 
 		*pRowid = newrow;
 	}
-	else if ((nData > 1) && (SQLITE3_CALL (sqlite3_value_type) (apData[0])==SQLITE_INTEGER)) {
+	else if (optype == 3) {
 		/* UPDATE */
 		gint i;
 
-		if (SQLITE3_CALL (sqlite3_value_int) (apData[0]) != SQLITE3_CALL (sqlite3_value_int) (apData[1])) {
-			/* argc>1 and argv[0]==argv[1]: rowid is imposed by SQLite which is not supported */
-			return SQLITE_READONLY;
-		}
+		
 		for (i = 2; i < (nData - 1); i++) {
 			GValue *value;
 			GType type;
@@ -954,7 +1316,8 @@ virtualUpdate (sqlite3_vtab *tab, int nData, sqlite3_value **apData, sqlite_int6
 			/*g_print ("%d => %s\n", i, SQLITE3_CALL (sqlite3_value_text) (apData [i]));*/
 			type = gda_column_get_g_type (gda_data_model_describe_column (vtable->td->real_model, i - 2));
 			value = gda_value_new_from_string ((const gchar*) SQLITE3_CALL (sqlite3_value_text) (apData [i]), type);
-			res = gda_data_model_set_value_at (vtable->td->real_model, i - 2, rowid, value, &error);
+			res = gda_data_model_set_value_at (vtable->td->real_model, i - 2, rowid,
+							   value, &error);
 			gda_value_free (value);
 			if (!res) {
 				g_print ("Error: %s\n", error && error->message ? error->message : "???");
@@ -979,7 +1342,7 @@ virtualUpdate (sqlite3_vtab *tab, int nData, sqlite3_value **apData, sqlite_int6
 static int
 virtualBegin (sqlite3_vtab *tab)
 {
-	TRACE (tab);
+	TRACE (tab, NULL);
 	/* no documentation currently available, don't do anything */
 	return SQLITE_OK;
 }
@@ -987,7 +1350,7 @@ virtualBegin (sqlite3_vtab *tab)
 static int
 virtualSync (G_GNUC_UNUSED sqlite3_vtab *tab)
 {
-	TRACE (tab);
+	TRACE (tab, NULL);
 	/* no documentation currently available, don't do anything */
 	return SQLITE_OK;
 }
@@ -995,23 +1358,30 @@ virtualSync (G_GNUC_UNUSED sqlite3_vtab *tab)
 static int
 virtualCommit (G_GNUC_UNUSED sqlite3_vtab *tab)
 {
-	TRACE (tab);
-	/* no documentation currently available, don't do anything */
+	VirtualTable *vtable = (VirtualTable *) tab;
+	TRACE (tab, NULL);
+
+	if (vtable->rowid_hash) {
+		g_hash_table_destroy (vtable->rowid_hash);
+		vtable->rowid_hash = NULL;
+		vtable->rowid_hash_model = NULL;
+	}
+
 	return SQLITE_OK;
 }
 
 static int
 virtualRollback (G_GNUC_UNUSED sqlite3_vtab *tab)
 {	
-	TRACE (tab);
+	TRACE (tab, NULL);
 	/* no documentation currently available, don't do anything */
 	return SQLITE_OK;
 }
 
 static int
-virtualRename(sqlite3_vtab *pVtab, const char *zNew)
+virtualRename (sqlite3_vtab *pVtab, G_GNUC_UNUSED const char *zNew)
 {
-	TRACE (pVtab);
+	TRACE (pVtab, NULL);
 	/* not yet analysed and implemented */
 	return SQLITE_OK;
 }
