@@ -1,5 +1,5 @@
 /* GDA Postgres provider
- * Copyright (C) 2008 The GNOME Foundation.
+ * Copyright (C) 2008 - 2010 The GNOME Foundation.
  *
  * AUTHORS:
  *      Vivien Malerba <malerba@gnome-db.org>
@@ -30,12 +30,11 @@
 #include "gda-postgres-provider.h"
 #include "gda-postgres-blob-op.h"
 #include "gda-postgres-util.h"
+#include <libgda/libgda-global-variables.h>
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
 
-extern gchar *gda_numeric_locale;
-
 #define _GDA_PSTMT(x) ((GdaPStmt*)(x))
 
 static void gda_postgres_recordset_class_init (GdaPostgresRecordsetClass *klass);
@@ -62,10 +61,10 @@ static gboolean gda_postgres_recordset_fetch_at (GdaDataSelect *model, GdaRow **
 
 /* static helper functions */
 static void make_point (GdaGeometricPoint *point, const gchar *value);
-static void set_value (GdaConnection *cnc, GValue *value, GType type, const gchar *thevalue, gint length);
+static void set_value (GdaConnection *cnc, GdaRow *row, GValue *value, GType type, const gchar *thevalue, gint length, GError **error);
 
-static void     set_prow_with_pg_res (GdaPostgresRecordset *imodel, GdaRow *prow, gint pg_res_rownum);
-static GdaRow *new_row_from_pg_res (GdaPostgresRecordset *imodel, gint pg_res_rownum);
+static void     set_prow_with_pg_res (GdaPostgresRecordset *imodel, GdaRow *prow, gint pg_res_rownum, GError **error);
+static GdaRow *new_row_from_pg_res (GdaPostgresRecordset *imodel, gint pg_res_rownum, GError **error);
 static gboolean row_is_in_current_pg_res (GdaPostgresRecordset *model, gint row);
 static gboolean fetch_next_chunk (GdaPostgresRecordset *model, gboolean *fetch_error, GError **error);
 static gboolean fetch_prev_chunk (GdaPostgresRecordset *model, gboolean *fetch_error, GError **error);
@@ -102,7 +101,7 @@ enum
  * Object init and finalize
  */
 static void
-gda_postgres_recordset_init (GdaPostgresRecordset *recset, GdaPostgresRecordsetClass *klass)
+gda_postgres_recordset_init (GdaPostgresRecordset *recset, G_GNUC_UNUSED GdaPostgresRecordsetClass *klass)
 {
 	g_return_if_fail (GDA_IS_POSTGRES_RECORDSET (recset));
 	recset->priv = g_new0 (GdaPostgresRecordsetPrivate, 1);
@@ -189,8 +188,9 @@ gda_postgres_recordset_set_property (GObject *object,
                 case PROP_CHUNCK_SIZE:
                         model->priv->chunk_size = g_value_get_int (value);
                         break;
-                default:
-                        break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
                 }
         }
 }
@@ -210,8 +210,9 @@ gda_postgres_recordset_get_property (GObject *object,
                 case PROP_CHUNCKS_READ:
                         g_value_set_int (value, model->priv->chunks_read);
                         break;
-                default:
-                        break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+			break;
                 }
         }
 }
@@ -236,7 +237,8 @@ gda_postgres_recordset_get_type (void)
 			NULL,
 			sizeof (GdaPostgresRecordset),
 			0,
-			(GInstanceInitFunc) gda_postgres_recordset_init
+			(GInstanceInitFunc) gda_postgres_recordset_init,
+			0
 		};
 		g_static_mutex_lock (&registering);
 		if (type == 0)
@@ -275,9 +277,11 @@ finish_prep_stmt_init (PostgresConnectionData *cdata, GdaPostgresPStmt *ps, PGre
 				if (col_types [i] > 0) {
 					if (col_types [i] == G_TYPE_NONE)
 						break;
-					if (i >= _GDA_PSTMT (ps)->ncols)
+					if (i >= _GDA_PSTMT (ps)->ncols) {
 						g_warning (_("Column %d out of range (0-%d), ignoring its specified type"), i,
 							   _GDA_PSTMT (ps)->ncols - 1);
+						break;
+					}
 					else
 						_GDA_PSTMT (ps)->types [i] = col_types [i];
 				}
@@ -295,7 +299,7 @@ finish_prep_stmt_init (PostgresConnectionData *cdata, GdaPostgresPStmt *ps, PGre
 			postgres_type = PQftype (pg_res, i);
 			gtype = _GDA_PSTMT (ps)->types [i];
 			if (gtype == 0) {
-				gtype = _gda_postgres_type_oid_to_gda (cdata, postgres_type);
+				gtype = _gda_postgres_type_oid_to_gda (cdata->cnc, cdata->reuseable, postgres_type);
 				_GDA_PSTMT (ps)->types [i] = gtype;
 			}
 			_GDA_PSTMT (ps)->types [i] = gtype;
@@ -432,7 +436,7 @@ gda_postgres_recordset_fetch_random (GdaDataSelect *model, GdaRow **prow, gint r
 		return FALSE;
 	}
 
-	*prow = new_row_from_pg_res (imodel, rownum);
+	*prow = new_row_from_pg_res (imodel, rownum, error);
 	gda_data_select_take_row (model, *prow, rownum);
 
 	if (model->nb_stored_rows == model->advertized_nrows) {
@@ -484,9 +488,9 @@ gda_postgres_recordset_fetch_next (GdaDataSelect *model, GdaRow **prow, gint row
 
 	if (row_is_in_current_pg_res (imodel, rownum)) {
 		if (imodel->priv->tmp_row)
-			set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf);
+			set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf, error);
 		else
-			imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+			imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 		*prow = imodel->priv->tmp_row;
 		return TRUE;
 	}
@@ -494,9 +498,9 @@ gda_postgres_recordset_fetch_next (GdaDataSelect *model, GdaRow **prow, gint row
 		gboolean fetch_error = FALSE;
 		if (fetch_next_chunk (imodel, &fetch_error, error)) {
 			if (imodel->priv->tmp_row)
-				set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf);
+				set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf, error);
 			else
-				imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+				imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 			*prow = imodel->priv->tmp_row;
 			return TRUE;
 		}
@@ -521,9 +525,9 @@ gda_postgres_recordset_fetch_prev (GdaDataSelect *model, GdaRow **prow, gint row
 
 	if (row_is_in_current_pg_res (imodel, rownum)) {
 		if (imodel->priv->tmp_row)
-			set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf);
+			set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf, error);
 		else
-			imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+			imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 		*prow = imodel->priv->tmp_row;
 		return TRUE;
 	}
@@ -531,9 +535,9 @@ gda_postgres_recordset_fetch_prev (GdaDataSelect *model, GdaRow **prow, gint row
 		gboolean fetch_error = FALSE;
 		if (fetch_prev_chunk (imodel, &fetch_error, error)) {
 			if (imodel->priv->tmp_row)
-				set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf);
+				set_prow_with_pg_res (imodel, imodel->priv->tmp_row, rownum - imodel->priv->pg_res_inf, error);
 			else
-				imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+				imodel->priv->tmp_row = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 			*prow = imodel->priv->tmp_row;
 			return TRUE;
 		}
@@ -562,14 +566,14 @@ gda_postgres_recordset_fetch_at (GdaDataSelect *model, GdaRow **prow, gint rownu
 	}
 
 	if (row_is_in_current_pg_res (imodel, rownum)) {
-		*prow = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+		*prow = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 		imodel->priv->tmp_row = *prow;
 		return TRUE;
 	}
 	else {
 		gboolean fetch_error = FALSE;
 		if (fetch_row_number_chunk (imodel, rownum, &fetch_error, error)) {
-			*prow = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf);
+			*prow = new_row_from_pg_res (imodel, rownum - imodel->priv->pg_res_inf, error);
 			imodel->priv->tmp_row = *prow;
 			return TRUE;
 		}
@@ -599,8 +603,8 @@ make_point (GdaGeometricPoint *point, const gchar *value)
 }
 
 static void
-set_value (GdaConnection *cnc, GValue *value, GType type,
-	   const gchar *thevalue, gint length)
+set_value (GdaConnection *cnc, GdaRow *row, GValue *value, GType type, const gchar *thevalue,
+	   G_GNUC_UNUSED gint length, GError **error)
 {
 	gda_value_reset_with_type (value, type);
 
@@ -610,17 +614,27 @@ set_value (GdaConnection *cnc, GValue *value, GType type,
 		g_value_set_string (value, thevalue);
 	else if (type == G_TYPE_INT)
 		g_value_set_int (value, atol (thevalue));
+	else if (type == G_TYPE_UINT)
+		g_value_set_uint (value, (guint) g_ascii_strtoull (thevalue, NULL, 10));
 	else if (type == G_TYPE_DATE) {
 		GDate date;
-		if (!gda_parse_iso8601_date (&date, thevalue)) 
-			g_warning (_("Invalid date '%s' (date format should be YYYY-MM-DD)"), thevalue);
+		if (!gda_parse_iso8601_date (&date, thevalue)) {
+			gda_row_invalidate_value (row, value);
+			g_set_error (error, GDA_SERVER_PROVIDER_ERROR,
+				     GDA_SERVER_PROVIDER_DATA_ERROR,
+				     _("Invalid date '%s' (date format should be YYYY-MM-DD)"), thevalue);
+		}
 		else
 			g_value_set_boxed (value, &date);
 	}
 	else if (type == GDA_TYPE_TIME) {
 		GdaTime timegda;
-		if (!gda_parse_iso8601_time (&timegda, thevalue)) 
-			g_warning (_("Invalid time '%s' (time format should be HH:MM:SS[.ms])"), thevalue);
+		if (!gda_parse_iso8601_time (&timegda, thevalue)) {
+			gda_row_invalidate_value (row, value); 
+			g_set_error (error, GDA_SERVER_PROVIDER_ERROR,
+				     GDA_SERVER_PROVIDER_DATA_ERROR,
+				     _("Invalid time '%s' (time format should be HH:MM:SS[.ms])"), thevalue);
+		}
 		else
 			gda_value_set_time (value, &timegda);
 	}
@@ -657,9 +671,13 @@ set_value (GdaConnection *cnc, GValue *value, GType type,
 	}
 	else if (type == GDA_TYPE_TIMESTAMP) {
 		GdaTimestamp timestamp;
-		if (! gda_parse_iso8601_timestamp (&timestamp, thevalue))
-			g_warning (_("Invalid timestamp '%s' (format should be YYYY-MM-DD HH:MM:SS[.ms])"), 
-				   thevalue);
+		if (! gda_parse_iso8601_timestamp (&timestamp, thevalue)) {
+			gda_row_invalidate_value (row, value);
+			g_set_error (error, GDA_SERVER_PROVIDER_ERROR,
+				     GDA_SERVER_PROVIDER_DATA_ERROR,
+				     _("Invalid timestamp '%s' (format should be YYYY-MM-DD HH:MM:SS[.ms])"), 
+				     thevalue);
+		}
 		else
 			gda_value_set_timestamp (value, &timestamp);
 	}
@@ -692,7 +710,7 @@ set_value (GdaConnection *cnc, GValue *value, GType type,
 	else if (type == G_TYPE_GTYPE)
 		g_value_set_gtype (value, gda_g_type_from_string (thevalue));
 	else {
-		g_warning ("Type %s not translated for value '%s' => set as string", g_type_name (type), thevalue);
+		/*g_warning ("Type %s not translated for value '%s' => set as string", g_type_name (type), thevalue);*/
 		gda_value_reset_with_type (value, G_TYPE_STRING);
 		g_value_set_string (value, thevalue);
 	}
@@ -709,7 +727,7 @@ row_is_in_current_pg_res (GdaPostgresRecordset *model, gint row)
 }
 
 static void
-set_prow_with_pg_res (GdaPostgresRecordset *imodel, GdaRow *prow, gint pg_res_rownum)
+set_prow_with_pg_res (GdaPostgresRecordset *imodel, GdaRow *prow, gint pg_res_rownum, GError **error)
 {
 	gchar *thevalue;
 	gint col;
@@ -720,20 +738,20 @@ set_prow_with_pg_res (GdaPostgresRecordset *imodel, GdaRow *prow, gint pg_res_ro
 			gda_value_set_null (gda_row_get_value (prow, col));
 		else
 			set_value (gda_data_select_get_connection ((GdaDataSelect*) imodel),
-				   gda_row_get_value (prow, col), 
+				   prow, gda_row_get_value (prow, col), 
 				   ((GdaDataSelect*) imodel)->prep_stmt->types [col], 
 				   thevalue, 
-				   PQgetlength (imodel->priv->pg_res, pg_res_rownum, col));
+				   PQgetlength (imodel->priv->pg_res, pg_res_rownum, col), error);
 	}
 }
 
 static GdaRow *
-new_row_from_pg_res (GdaPostgresRecordset *imodel, gint pg_res_rownum)
+new_row_from_pg_res (GdaPostgresRecordset *imodel, gint pg_res_rownum, GError **error)
 {
 	GdaRow *prow;
 
 	prow = gda_row_new (((GdaDataSelect*) imodel)->prep_stmt->ncols);
-	set_prow_with_pg_res (imodel, prow, pg_res_rownum);
+	set_prow_with_pg_res (imodel, prow, pg_res_rownum, error);
 	return prow;
 }
 
